<?php

namespace Oro\Bundle\OAuth2ServerBundle\Controller;

use Oro\Bundle\EntityBundle\Tools\EntityRoutingHelper;
use Oro\Bundle\FormBundle\Model\UpdateHandlerFacade;
use Oro\Bundle\OAuth2ServerBundle\Entity\Client;
use Oro\Bundle\OAuth2ServerBundle\Entity\Manager\ClientManager;
use Oro\Bundle\OAuth2ServerBundle\Form\Type\ClientType;
use Oro\Bundle\OAuth2ServerBundle\Form\Type\SystemClientType;
use Oro\Bundle\OAuth2ServerBundle\Security\EncryptionKeysExistenceChecker;
use Oro\Bundle\SecurityBundle\Annotation\CsrfProtection;
use Sensio\Bundle\FrameworkExtraBundle\Configuration\Route;
use Sensio\Bundle\FrameworkExtraBundle\Configuration\Template;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\Form\FormFactoryInterface;
use Symfony\Component\Form\FormInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Translation\TranslatorInterface;

/**
 * This controller covers widget-related functionality for OAuth 2.0 Client entity.
 */
class ClientController extends AbstractController
{
    /** @var array */
    private $supportedGrantTypes;

    /** @var array */
    private $supportedClientTypes = [
        'storefront' => ['isFrontend' => true],
        'backoffice' => ['isFrontend' => false]
    ];

    /**
     * @param array $supportedGrantTypes
     */
    public function __construct(array $supportedGrantTypes)
    {
        $this->supportedGrantTypes = $supportedGrantTypes;
    }

    /**
     * {@inheritdoc}
     */
    public static function getSubscribedServices()
    {
        return array_merge(parent::getSubscribedServices(), [
            EntityRoutingHelper::class,
            UpdateHandlerFacade::class,
            ClientManager::class,
            FormFactoryInterface::class,
            TranslatorInterface::class,
            EncryptionKeysExistenceChecker::class
        ]);
    }

    /**
     * @Route(
     *     "/",
     *     name="oro_oauth2_index",
     *     defaults={"type": "backoffice"}
     * )
     * @Route(
     *     "/storefront",
     *     name="oro_oauth2_storefront_index",
     *     defaults={"type": "storefront"}
     * )
     * @Template("OroOAuth2ServerBundle:Client:index.html.twig")
     *
     * @param string $type
     *
     * @return array
     */
    public function indexAction(string $type): array
    {
        $this->checkTypeEnabled($type);

        if (!$this->getClientManager()->isViewGranted()) {
            throw $this->createAccessDeniedException();
        }

        return [
            'isFrontend' => $this->supportedClientTypes[$type]['isFrontend'],
            'encryptionKeysExist' => $this->isEncryptionKeysExist()
        ];
    }

    /**
     * @Route(
     *     "/{id}",
     *     name="oro_oauth2_view",
     *     requirements={"id"="\d+"},
     *     defaults={"type": "backoffice"}
     * )
     * @Route(
     *     "/storefront/{id}",
     *     name="oro_oauth2_storefront_view",
     *     requirements={"id"="\d+"},
     *     defaults={"type": "storefront"}
     * )
     *
     * @Template("OroOAuth2ServerBundle:Client:view.html.twig")
     *
     * @param Client $entity
     * @param string $type
     *
     * @return mixed
     */
    public function viewAction(Client $entity, string $type): array
    {
        $this->checkTypeEnabled($type);
        $this->checkClientApplicableForType($entity, $type);

        if (!$this->getClientManager()->isViewGranted($entity)) {
            throw $this->createAccessDeniedException();
        }

        $user = null;
        if ($entity->getOwnerEntityId()) {
            $user = $this->getDoctrine()->getRepository($entity->getOwnerEntityClass())->find(
                $entity->getOwnerEntityId()
            );
        }

        return ['entity' => $entity, 'user' => $user, 'encryptionKeysExist' => $this->isEncryptionKeysExist()];
    }

    /**
     * @Route(
     *     "/create",
     *     name="oro_oauth2_create",
     *     defaults={"type": "backoffice"},
     *     methods={"GET", "POST"}
     * )
     * @Route(
     *     "/create/storefront",
     *     name="oro_oauth2_storefront_create",
     *     defaults={"type": "storefront"},
     *     methods={"GET", "POST"}
     * )
     *
     * @Template("OroOAuth2ServerBundle:Client:create.html.twig")
     *
     * @param Request $request
     * @param string  $type
     *
     * @return mixed
     */
    public function createAction(Request $request, string $type)
    {
        $this->checkTypeEnabled($type);
        if (!$this->getClientManager()->isCreationGranted()) {
            throw $this->createAccessDeniedException();
        }

        $entity = new Client();
        $entity->setFrontend($this->supportedClientTypes[$type]['isFrontend']);

        $response = $this->update(
            $request,
            $entity,
            true,
            $this->supportedGrantTypes,
            $this->getTranslator()->trans('oro.oauth2server.client.created_message')
        );

        // change the Redirect response to custom template render to be able to show client secret information.
        if ($response instanceof RedirectResponse) {
            $response = $this->render(
                'OroOAuth2ServerBundle:Client:createResult.html.twig',
                [
                    'entity' => $entity,
                    'type'   => $type
                ]
            );
        }

        return $response;
    }

    /**
     * @Route(
     *     "/update/{id}",
     *     name="oro_oauth2_update",
     *     defaults={"type": "backoffice"},
     *     methods={"GET", "POST"},
     *     requirements={"id"="\d+"}
     * )
     * @Route(
     *     "/update/storefront/{id}",
     *     name="oro_oauth2_storefront_update",
     *     defaults={"type": "storefront"},
     *     methods={"GET", "POST"},
     *     requirements={"id"="\d+"}
     * )
     *
     * @Template("OroOAuth2ServerBundle:Client:update.html.twig")
     *
     * @param Request $request
     * @param Client  $entity
     * @param string  $type
     *
     * @return mixed
     */
    public function updateAction(Request $request, Client $entity, string $type)
    {
        $this->checkTypeEnabled($type);
        $this->checkClientApplicableForType($entity, $type);
        $this->checkModificationAccess($entity);

        return $this->update(
            $request,
            $entity,
            true,
            $this->supportedGrantTypes,
            $this->getTranslator()->trans('oro.oauth2server.client.updated_message')
        );
    }

    /**
     * @Route(
     *     "/create/client",
     *     name="oro_oauth2_server_client_create",
     *     methods={"GET", "POST"}
     * )
     * @Template("OroOAuth2ServerBundle:Client:create.html.twig")
     *
     * @param Request $request
     *
     * @return array
     */
    public function createClientAction(Request $request)
    {
        if (!$this->getClientManager()->isCreationGranted()) {
            throw $this->createAccessDeniedException();
        }

        $entity = new Client();
        $entityRoutingHelper = $this->get(EntityRoutingHelper::class);
        $ownerEntityClass = $entityRoutingHelper->getEntityClassName($request);
        if ($ownerEntityClass) {
            $ownerEntityId = (int)$entityRoutingHelper->getEntityId($request);
            $entity->setOwnerEntity($ownerEntityClass, $ownerEntityId);
        }

        return $this->update($request, $entity);
    }

    /**
     * @Route(
     *     "/update/client/{id}",
     *     name="oro_oauth2_server_client_update",
     *     methods={"GET", "POST"},
     *     requirements={"id"="\d+"}
     * )
     * @Template("OroOAuth2ServerBundle:Client:update.html.twig")
     *
     * @param Request $request
     * @param Client  $entity
     *
     * @return array
     */
    public function updateClientAction(Request $request, Client $entity)
    {
        $this->checkModificationAccess($entity);

        return $this->update($request, $entity);
    }

    /**
     * @Route(
     *     "/delete/{id}",
     *     name="oro_oauth2_server_client_delete",
     *     methods={"DELETE"},
     *     requirements={"id"="\d+"}
     * )
     * @CsrfProtection()
     *
     * @param Client $entity
     *
     * @return Response
     */
    public function deleteAction(Client $entity): Response
    {
        if (!$this->getClientManager()->isDeletionGranted($entity)) {
            throw $this->createAccessDeniedException();
        }

        $this->getClientManager()->deleteClient($entity);

        return new JsonResponse(['successful' => true]);
    }

    /**
     * @Route(
     *     "/activate/{id}",
     *     name="oro_oauth2_server_client_activate",
     *     methods={"POST"},
     *     requirements={"id"="\d+"}
     * )
     * @CsrfProtection()
     *
     * @param Client $entity
     *
     * @return Response
     */
    public function activateAction(Client $entity): Response
    {
        $this->checkModificationAccess($entity);

        $this->getClientManager()->activateClient($entity);

        return new JsonResponse(['successful' => true]);
    }

    /**
     * @Route(
     *     "/deactivate/{id}",
     *     name="oro_oauth2_server_client_deactivate",
     *     methods={"POST"},
     *     requirements={"id"="\d+"}
     * )
     * @CsrfProtection()
     *
     * @param Client $entity
     *
     * @return Response
     */
    public function deactivateAction(Client $entity): Response
    {
        $this->checkModificationAccess($entity);

        $this->getClientManager()->deactivateClient($entity);

        return new JsonResponse(['successful' => true]);
    }

    /**
     * @param Request $request
     * @param Client  $entity
     * @param bool    $isSystemOAuthApplication
     * @param array   $grantTypes
     * @param string  $message
     *
     * @return array
     */
    private function update(
        Request $request,
        Client $entity,
        $isSystemOAuthApplication = false,
        $grantTypes = ['client_credentials'],
        $message = null
    ) {
        return $this->get(UpdateHandlerFacade::class)->update(
            $entity,
            $this->getForm($entity, $isSystemOAuthApplication, $grantTypes),
            $message,
            $request,
            null,
            $this->getFormTemplateDataProvider($isSystemOAuthApplication)
        );
    }

    /**
     * @param Client $entity
     */
    private function checkModificationAccess(Client $entity): void
    {
        if (!$this->getClientManager()->isModificationGranted($entity)) {
            throw $this->createAccessDeniedException();
        }
    }

    /**
     * @return ClientManager
     */
    private function getClientManager(): ClientManager
    {
        return $this->get(ClientManager::class);
    }

    /**
     * @param Client $entity
     * @param bool   $isSystemOAuthApplication
     * @param array  $grantTypes
     *
     * @return FormInterface
     */
    private function getForm(
        Client $entity,
        $isSystemOAuthApplication = false,
        $grantTypes = ['client_credentials']
    ): FormInterface {
        return $this->get(FormFactoryInterface::class)
            ->createNamed(
                'oro_oauth2_client',
                $isSystemOAuthApplication ? SystemClientType::class : ClientType::class,
                $entity,
                [
                    'grant_types' => $grantTypes,
                    'show_grants' => $isSystemOAuthApplication
                ]
            );
    }

    /**
     * @return callable
     */
    private function getFormTemplateDataProvider($isSystemOAuthApplication = false): callable
    {
        return function (Client $entity, FormInterface $form) use ($isSystemOAuthApplication) {
            if ($isSystemOAuthApplication) {
                $formAction = null === $entity->getId()
                    ? $this->generateUrl($entity->isFrontend() ? 'oro_oauth2_storefront_create' : 'oro_oauth2_create')
                    : $this->generateUrl(
                        $entity->isFrontend() ? 'oro_oauth2_storefront_update' : 'oro_oauth2_update',
                        ['id' => $entity->getId()]
                    );
            } else {
                $formAction = null === $entity->getId()
                    ? $this->generateUrl('oro_oauth2_server_client_create')
                    : $this->generateUrl('oro_oauth2_server_client_update', ['id' => $entity->getId()]);
            }

            return [
                'entity'     => $entity,
                'form'       => $form->createView(),
                'formAction' => $formAction
            ];
        };
    }

    /**
     * @return TranslatorInterface
     */
    private function getTranslator()
    {
        return $this->container->get(TranslatorInterface::class);
    }

    /**
     * @param string $type
     */
    private function checkTypeEnabled(string $type): void
    {
        if ($type === 'storefront' && !class_exists('Oro\Bundle\CustomerBundle\OroCustomerBundle')) {
            throw $this->createNotFoundException();
        }
    }

    /**
     * @param Client $client
     * @param string $type
     */
    private function checkClientApplicableForType(Client $client, string $type): void
    {
        $isFrontend = $type === 'storefront';
        if ($client->isFrontend() !== $isFrontend) {
            throw $this->createNotFoundException();
        }
    }

    /**
     * @return bool
     */
    private function isEncryptionKeysExist()
    {
        $encryptionKeysExistenceChecker = $this->get(EncryptionKeysExistenceChecker::class);

        return $encryptionKeysExistenceChecker->isPrivateKeyExist()
                && $encryptionKeysExistenceChecker->isPublicKeyExist();
    }
}
